/*
  JavaScript utilities for PmWiki
  (c) 2009-2026 Petko Yotov www.pmwiki.org/petko
  based on PmWiki addons DeObMail, AutoTOC and Ape
  licensed GNU GPLv2 or any more recent version released by the FSF.

  libsortable() "Sortable tables" adapted for PmWiki from
  a Public Domain event listener by github.com/tofsjonas
*/

(async function(__script__){
  
  const { echo, err, Now, qs, qsa, dqs, dqsa, PHSC, dce,
    css, min, max, minmax, floor, floatval, intval,
    abs, zpad, getLS, setLS, jP, getConf, aE, dE, tap,
    adjbb, adjbe, adjab, adjae, posy, mkDrag,
    gcs, fe4j, fe4b, fe4t, fe4, fe4p, ready } = PmLib;
    
  var Config = getConf(__script__);
  
  if(Config.rediquiet) {
    var url = location.href.replace(/\?from=[^?&#]*[&]quiet=1/, '');
    if(url != location.href) 
      history.replaceState(null, null, url);
  }

  var wikitext;

  function PmXMail() {
    var els = dqsa('span._pmXmail');
    var LinkFmt = '<a href="%u" class="mail">%t</a>';

    for(var i=0; i<els.length; i++) {
      var x = qs(els[i], 'span._t');
      var txt = cb_mail(x.innerHTML);
      var y = qs(els[i], 'span._m');
      var url = cb_mail(y.innerHTML.replace(/^ *-&gt; */, ''));

      if(!url) url = 'mailto:'+txt.replace(/^mailto:/, '');

      url = url.replace(/"/g, '%22').replace(/'/g, '%27');
      var html = LinkFmt.replace(/%u/g, url).replace(/%t/g, txt);
      els[i].innerHTML = html;
    }
  }
  function cb_mail(x){
    return x.replace( /<span class=(['"]?)_d\1>[^<]+<\/span>/ig, '.')
      .replace( /<span class=(['"]?)_a\1>[^<]+<\/span>/ig, '@');
  }

  function is_toc_heading(el) {
    if(el.offsetParent === null) {return false;}  // hidden
    if(el.closest('.notoc,.markup2')) {return false;} 
    return true;
  }
  
  function any_id(h) {
    if(h.id) {return h.id;} // %id=anchor%
    var a = qs(h, 'a[id]'); // inline [[#anchor]]
    if(a && a.id) {return a.id;}
    var prev = h.previousElementSibling;
    if(prev) { // [[#anchor]] before !!heading
      var a = qsa(prev, 'a[id]');
      if(a.length) {
        last = a[a.length-1];
        if(last.id && ! last.nextElementSibling) {
          var atop = posy(last) + last.offsetHeight;
          var htop = posy(h);
          if( abs(htop-atop)<20 ) {
            h.appendChild(last);
            return last.id;
          }
        }
      }
    }
    return false;
  }

  
  function autotoc() {
    if(dqs('.noPmTOC')) { return; } // (:notoc:) in page
    var dtoc = Config.pmtoc;
    if(! dtoc.Enable || !dtoc.MaxLevel) { return; } // disabled

    if(dtoc.NumberedHeadings)  {
      var specs = dtoc.NumberedHeadings.toString().split(/\./g);
      for(var i=0; i<specs.length; i++) {
        if(specs[i].match(/^[1AI]$/i)) numheadspec[i] = ''+specs[i];
        else numheadspec[i] = '';
      }
      if(!dtoc.MinNumberedHeadings) dtoc.MinNumberedHeadings = 2;
    }

    var query = [];
    for(var i=1; i<=dtoc.MaxLevel; i++) {
      query.push('h'+i);
    }
    if(dtoc.EnableQMarkup) query.push('p.question');
    var pageheadings = qsa(wikitext, query.join(','));
    if(!pageheadings.length) { return; }

    var toc_headings = [ ];
    var minlevel = 1000, hcache = [ ];
    for(var i=0; i<pageheadings.length; i++) {
      var h = pageheadings[i];
      if(! is_toc_heading(h)) {continue;}
      toc_headings.push(h);
    }
    if(! toc_headings.length) return;

    var tocdiv = dqs('.PmTOCdiv');
    var shouldmaketoc = ( tocdiv || (toc_headings.length >= dtoc.MinNumber && dtoc.MinNumber != -1)) ? 1:0;
    var shouldnumheadings = (dtoc.NumberedHeadings && toc_headings.length >= dtoc.MinNumberedHeadings) ? 1:0;
    if(!shouldnumheadings && !shouldmaketoc) return;

    for(var i=0; i<toc_headings.length; i++) {
      var h = toc_headings[i];
      var level = floatval(h.tagName.substring(1));
      if(! level) level = 6;
      minlevel = min(minlevel, level);
      var id = any_id(h);
      hcache.push([h, level, id]);
    }
    
    var blink = false;
    if(dtoc.EnableBacklinks) {
      let bltip = dtoc.EnableBacklinks === 1 ? dtoc.contents:dtoc.EnableBacklinks;
      blink = dce('a', { href:"#_toc", className:"back-arrow", title:bltip });
    }

    prevlevel = 0;
    var html = '';
    var toclinks = [];
    let frag = dce('template');
    for(var i=0; i<hcache.length; i++) {
      var hc = hcache[i];
      var actual_level = hc[1] - minlevel;

      var currnb = numberheadings(actual_level);
      if(! hc[2]) {
        hc[2] = 'toc-'+currnb.id.replace(/\.+$/g, '');
        hc[0].id = hc[2];
      }
      
      if(shouldmaketoc) {
        frag = frag.cloneNode();
        frag.innerHTML = hc[0].innerHTML;
        
        let se = qsa(frag.content, '.sectionedit, .rmTOC');
        for(let se1 of se) se1.remove();
        let ke = qsa(frag.content, '.kpTOC');
        for(let ki=0; ki<ke.length; ki++) {
          let ker = dce('span', {textContent:`[kT${ki}kT]`});
          adjbb(ke[ki], ker);
          ke[ki].remove();
        }
        let txt = PHSC(frag.content.textContent.trim());
        if(ke.length) txt = txt.replace(/\[kT(\d+)kT\]/g, function(a, a1){
          return ke[intval(a1)].outerHTML;
        });
        
        var toclink = dce('a', {
          className:'pmtoc-indent' + actual_level,
          href: '#' + hc[2],
          innerHTML: txt
        });
        toclinks.push(toclink);
        if(blink) adjbe(hc[0], blink.cloneNode());
      }
      if(shouldnumheadings) {
        let alevel1 = actual_level+1;
        nhtml = `<span class="pmnumwrap pmnumwrap${alevel1}">${currnb.num} </span>`;
        adjab(hc[0], nhtml);
        if(shouldmaketoc) adjab(toclink, nhtml);
      }
    }

    if(! shouldmaketoc) return;
 
    var details = dce('details', {
      id: '_toc',
      className: 'PmTOCdiv frame'
    });
    if(! getLS('closeTOC', dtoc.Closed)) details.open = true;
 
    var sum = dce('summary', {textContent: dtoc.contents});
    details.appendChild(sum);
    
    var nav = dce('nav', { className:'PmTOCtable' });
    
    for(var a of toclinks) nav.appendChild(a);
    
    details.appendChild(nav);

    if(tocdiv) {
      adjbb(tocdiv, details);
      tocdiv.remove();
    }
    else {
      if(dtoc.ParentElement && dqs(dtoc.ParentElement)) {
        adjab(dqs(dtoc.ParentElement), details);
      }
      else {
        var x1 = hcache[0][0];
        var x0 = x1.previousElementSibling;
        if(x0 && x0.matches('.sectionedit.sectionhead')) x1 = x0;
        adjbb(x1, details);
      }
    }
    
    aE(details, 'toggle', function(e){
      setLS('closeTOC', this.open ? '' : 1);
    });

    var hh = location.hash;
    if(hh.length>1) {
      var cc = dqs(hh);
      if(cc) cc.scrollIntoView();
    }
  }

  var numhead = [0, 0, 0, 0, 0, 0, 0];
  var numheadspec = '1 1 1 1 1 1 1'.split(/ /g);
  function numhead_alpha(n, upper) {
    if(!n) return '_';
    var alpha = '', mod, start = upper=='A' ? 65 : 97;
    while (n>0) {
      mod = (n-1)%26;
      alpha = String.fromCharCode(start + mod) + '' + alpha;
      n = (n-mod)/26 | 0;
    }
    return alpha;
  }
  function numhead_roman(n, upper) {
    if(!n) return '_';
    // partially based on http://blog.stevenlevithan.com/?p=65#comment-16107
    var lst = [ [1000,'M'], [900,'CM'], [500,'D'], [400,'CD'], [100,'C'], [90,'XC'],
      [50,'L'], [40,'XL'], [10,'X'], [9,'IX'], [5,'V'], [4,'IV'], [1,'I'] ];
    var roman = '';
    for(var i=0; i<lst.length; i++) {
      while(n>=lst[i][0]) {
        roman += lst[i][1];
        n -= lst[i][0];
      }
    }
    return (upper == 'I') ? roman : roman.toLowerCase();
  }

  function numberheadings(n) {
    if(n<numhead[6]) for(var j=numhead[6]; j>n; j--) numhead[j]=0;
    numhead[6]=n;
    numhead[n]++;
    var qq = '';
    var hh = '';
    for (var j=0; j<=n; j++) {
      let j1 = j+1;
      let curr = numhead[j];
      hh += curr+".";
      let currspec = numheadspec[j];
      if(!currspec.match(/[1ai]/i)) continue;
      if(currspec.match(/a/i)) { curr = numhead_alpha(curr, currspec); }
      else if(currspec.match(/i/i)) { curr = numhead_roman(curr, currspec); }
      qq+= `<span class="pmnum${j1}">${curr}.</span>`;
      
    }
    return {id:hh.replace(/\.$/, ''), num:qq};
  }

  function makesortable() {
    if(! floatval(Config.sortable)) return;
    var tables = dqsa('table.sortable,table.sortable-footer');
    for(var i=0; i<tables.length; i++) {
      if(qs(tables[i], '[rowspan],[colspan]')) {
        tables[i].classList.replace('sortable', 'unsortable');
        tables[i].classList.replace('sortable-footer', 'unsortable-footer');
        continue;
      }
      var rows = qsa(tables[i], 'tr');
      // non-pmwiki-core table, already ready
      if(! qs(tables[i], 'thead')) {
        var thead = dce('thead');
        adjab(tables[i], thead);
        thead.appendChild(rows[0]);
      }
      var tbody = qs(tables[i], 'tbody');
      if(! tbody) {
        tbody = dce('tbody');
        tables[i].appendChild(tbody);
        for(var r=1; r<rows.length; r++) tbody.appendChild(rows[r]);
      }
      if(tables[i].classList.contains('sortable-footer')) {
        tables[i].classList.add('sortable');
        var tfoot = dce('tfoot');
        tables[i].appendChild(tfoot);
        tfoot.appendChild(rows[rows.length-1]);
      }
      mkdatasort(rows);
    }
    libsortable();
  }
  function mkdatasort(rows) {
    var hcells = qsa(rows[0], 'th,td');
    var specialsort = [], span;
    for(var i=0; i<hcells.length; i++) {
      sortspan = qs(hcells[i], '.sort-number,.sort-number-us,.sort-date,.sort-time');
      specialsort[i] = sortspan? sortspan.className : false;
    }
    if(! specialsort.length) return;
    for(var i=1; i<rows.length; i++) {
      var cells = qsa(rows[i], 'td,th');
      var k = 0;
      for(var j=0; j<cells.length && j<specialsort.length; j++) {
        if(! specialsort[j]) continue;
        var t = cells[j].innerText, ds = '';
        if(specialsort[j] == 'sort-number-us') {ds = t.replace(/[^-.\d]+/g, ''); }
        else if(specialsort[j] == 'sort-number') {ds = t.replace(/[^-,\d]+/g, '').replace(/,/g, '.'); }
        else if(specialsort[j] == 'sort-date') {ds = new Date(t).getTime(); }
        else if(specialsort[j] == 'sort-time') {
          var time = qs(cells[j],'time[datetime]');
          if(time) ds = time.datetime;
        }
        if(ds) cells[j].dataset.sort = ds;
      }
    }
  }
  function libsortable(){
    // adapted from Public Domain code by github.com/tofsjonas
    // updated to keep any event handlers in the table
    function getValue(obj) {
      obj = obj.cells[column_index];
      return obj.dataset.sort || obj.innerText;
    }
    function reclassify(element, cname) {
      element.classList.remove('dir-u', 'dir-d');
      if(cname) element.classList.add(cname);
    }
    var column_index;
    document.addEventListener('click', function(e) {
      if(e.target.closest('a')) return; // links
      var element = e.target.closest('th');
      if(! element) return;
      var table = element.closest('table');
      if(!table.classList.contains('sortable')) return;

      var prev = qsa(table, 'th.dir-d, th.dir-u');
      for (let cell of prev) if(cell !== element) reclassify(cell, '');

      var cells = [...element.parentNode.cells];
      column_index = cells.indexOf(element);

      var cname = 'dir-d', reverse = false;
      if(element.classList.contains(cname)) {
        cname = 'dir-u';
        reverse = true;
      }
      reclassify(element, cname);
      var tbody = table.tBodies[0];
      var rows = [...tbody.rows];

      rows.sort(function(x, y) {
        var a = getValue(reverse? y:x),
            b = getValue(reverse? x:y);
        var c = a.localeCompare(b, undefined, {numeric: true, sensitivity: 'base'}),
            d = a - b;
        return isNaN(d) ? c : d;
      });
      for (let row of rows) tbody.appendChild(row);

    });
  }

  function copy_pre() {
    if(!navigator.clipboard) return;
    var copytext = Config.copycode
    if(!copytext) return;
    var pres = dqsa('#wikitext pre');
    var btn = "<span class='pmcopycode frame rfloat' title='"+PHSC(copytext)+"'></span>"
    for(var i=0; i<pres.length; i++) {
      adjab(pres[i], btn);
    }
    tap('.pmcopycode', function(e){
      var cc = this;
      var pre = cc.parentNode;
      var tc = pre.textContent;
      navigator.clipboard.writeText(tc);
      cc.classList.add('copied');
      setTimeout(function(){cc.classList.remove('copied');}, 3000);
    });
  }

  function highlight_pre() {
    if(!Config.highlight) return;
    if(typeof hljs == 'undefined') return;
    
    var x = dqsa('.highlight,.hlt');
    
    for(var i=0; i<x.length; i++) {
      if(x[i].className.match(/(^| )(pm|pmwiki)( |$)/)) { continue;} // core highlighter
      var pre = [...qsa(x[i],'pre,code')];
      var n = x[i].nextElementSibling;
      if(n && n.tagName == 'PRE') pre.push(n);
      for(var j=0; j<pre.length; j++) {
        pre[j].className += ' ' + x[i].className;
        var varlinks = qsa(pre[j], 'a.varlink');
        var vararray = {};
        for(var v=0; v<varlinks.length; v++) {
          vararray[varlinks[v].textContent] = varlinks[v].href;
        }
        
        if(pre[j].children) pre[j].textContent = pre[j].textContent;

        hljs.highlightElement(pre[j]);
        var hlvars = qsa(pre[j], 'span.hljs-variable');
        for(var v=0; v<hlvars.length; v++) {
          var hlvar = hlvars[v].textContent;
          if(vararray.hasOwnProperty(hlvar)) {
            hlvars[v].innerHTML = '<a class="varlink" href="'+vararray[hlvar]+'">'+hlvar+'</a>';
          }
        }
      }
    }
    var varlinks = dqsa('code.varlink');
    for(i=0; i<varlinks.length; i++) varlinks[i].classList.add('hljs-variable');
  }
  
  var ltmode, daymonth, pagename;
  function fmtLocalTime(stamp) { // TODO: future dates
    var d = new Date(stamp*1000);
    var tooltip = PHSC(d.toLocaleString());
    if(ltmode == 2)
      return [tooltip];
    if(Now-d < 24*3600000) 
      return [zpad(d.getHours()) + ':'+ zpad(d.getMinutes()), tooltip];
    var D = zpad(d.getDate()), M = zpad(d.getMonth()+1);
    var thedate = daymonth.replace(/%d/, D).replace(/%m/, M);
    if(Now-d < 334*24*3600000) return [thedate, tooltip];
 
    if(ltmode == 1)
      return [thedate + '/' + d.getFullYear(), tooltip];
    
    if(ltmode == 3)
      return [M + "'" + zpad(d.getFullYear()%100), tooltip];
  }
  
  function localTimes() {
    ltmode = floatval(Config.localtimes);
    if(! ltmode) return;
    if(ltmode>=11) {
      var days = floor(ltmode/10);
      ltmode = ltmode%10;
    }
    else var days = 3;
 
    pagename = Config.fullname;
    var seenstamp = getLS('seenstamp', {}, true);
    var previous = seenstamp[pagename];
    
    var times = dqsa('time[datetime]');

    daymonth =  new Date(2021, 11, 26, 17)
      .toLocaleDateString().match(/26.*12/)? '%d/%m': '%m/%d';
    
    var h72 = Now.getTime()/1000-days*24*3600;
    
    for(var i=0; i<times.length; i++) {
      var itemdate = new Date(times[i].dateTime);
      var stamp = floor(itemdate.getTime()/1000);
      
      var li = times[i].closest('li');
      if(!li || !li.innerHTML.match(/<\/a>  \. \. \. /)) {
        var x = fmtLocalTime(stamp);
        times[i].innerHTML = x[0];
        times[i].title = x[1] ? x[1]: itemdate.textContent;
        continue;
      }
      var link = qs(li, 'a');
      if(link.className.match(/createlinktext|wikilink|selflink/)) {
        var diff = link.href + '?action=diff#diff' + stamp;
        if(stamp >= h72) {
          var h = link.href + '?action=diff&amp;fmt=rclist';
          adjbe(li, ' <input type="button" class="inputbutton rcplus" data-url="'+h+'" value="+">');
        }
      }
      // recent uploads, other? we want to know when the link becomes "visited"
      else diff = link.href + '#diff' + stamp; 
      times[i].innerHTML = '<a href="'+diff+'">'+times[i].innerHTML+'</a>';
    }

    var difflinks = dqsa('a[href*="#diff"]'), diffcnt = 0;
    for(var i=0; i<difflinks.length; i++) {
      var link = difflinks[i];
      if(link.hostname != location.hostname) continue;
      var a = link.href.match(/[#]diff(\d+)$/);
      if(!a) continue;
      diffcnt++;
      stamp = parseInt(a[1]);
      var x = fmtLocalTime(stamp);
      
      link.innerHTML = x[0];
      link.title = x[1] ? x[1]: link.textContent;
      
      var par = link.closest('li');
      if(!par) continue;
      adjab(par, link);
      adjae(link, "&nbsp;&nbsp;");
      if(previous && stamp>previous) par.classList.add('rcnew');
    }
    if(!diffcnt) return;
    var pagetitle = dqs('#wikititle h1, h1.pagetitle');
    if(pagetitle) {
      var time = zpad(Now.getHours()) + ':'+ zpad(Now.getMinutes());
      adjbe(pagetitle, ' <span class="rcreload" title="Click to reload">'+time+'</span>');
      tap('.rcreload', function(){location.reload();});
    }
    aE('.rcnew', 'mouseup', function(e){
      if(e.which == 2) this.classList.remove('rcnew');
    });
    tap('.rcplus', function(e){
      var plus = this;
      if(plus.value == '-') {
        var outdents = qsa(plus.closest('li'), 'p.outdent');
        for(var o of outdents) o.remove();
        plus.value = "+";
        return;
      }
      css(plus, {width:gcs(plus, 'width')});
      plus.value = "~";
      plus.disabled = true;
      var basehref = plus.dataset.url.replace(/&fmt=rclist/, '#diff')
        .replace(/[&]/g, '&amp;');
      var fmt = '<p class="outdent"><a href="'+basehref+'%d" title="%T">%t</a> %s</p>\n';
      fetch(plus.dataset.url)
      .then(function(resp){return resp.text();})
      .then(function(text){
        var lines = text.split(/\n/g);
        var out = '';
        for(var i=0; i<lines.length; i++) {
          a = lines[i].match(/^(\d+):(.*)$/);
          if(!a) continue;
          var time = fmtLocalTime(floatval(a[1]));
          out += fmt.replace(/%d/, a[1]).replace(/%T/, time[1])
            .replace(/%t/, time[0]).replace(/%s/, a[2]);
        }
        if(out) {
          adjae(plus, out);
          plus.value = "-";
          plus.disabled = false;
        }
      })
      .catch(err);
    });
    if(dqs('form[name="authform"]') || location.href.match(/action=/)) return;
    seenstamp[pagename] = floor(Now.getTime()/1000);
    setLS('seenstamp', seenstamp);
  }
  
  function confirmForms() {
    var forms = dqsa('form[data-pmconfirm]');
    aE('form[data-pmconfirm]', 'submit', function(e){
      if(!confirm(this.dataset.pmconfirm)) {
        e.preventDefault();
        e.stopPropagation();
      }
    });
    
    tap('button[data-pmconfirm],input[data-pmconfirm]', function(e){
      if(this.tagName == 'INPUT' && !this.type.match(/^(submit|reset|button)$/i)) return;
      if(!confirm(this.dataset.pmconfirm)) {
        e.preventDefault();
        e.stopPropagation();
      }
    });
  }
  
  function labelForms() {
    var elems = dqsa('input[type="color"][data-labelvalue="1"],input[type="range"][data-labelvalue="1"]');
    if(!elems.length) return;
    for(var el of elems) {
      var v = dce('var', {className: 'labelvalue'});
      v.dataset.value = el.value;
      adjae(el, v);
      el._labelvalue = v;
    }
    aE(elems, 'input', function(e) {
      this._labelvalue.dataset.value = this.value;
    });
  }
  
  function inittoggle() {
    var tnext = Config.toggle;
    if(! tnext) { return; }
    var x = dqsa(tnext);
    if(! x.length) return;
    for(var i=0; i<x.length; i++) togglenext(x[i]);
    tap(tnext, togglenext);
    tap('.pmtoggleall', toggleall);
  }
  function togglenext(z) {
    var el = z.type == 'click' ? this : z;
    var attr = el.dataset.pmtoggle=='closed' ? 'open' : 'closed';
    el.dataset.pmtoggle = attr;
  }
  function toggleall(){
    var curr = this.dataset.pmtoggleall;
    if(!curr) curr = 'closed';
    var toggles = dqsa('*[data-pmtoggle="'+curr+'"]');
    var next = curr=='closed' ? 'open' : 'closed';
    for(var i=0; i<toggles.length; i++) {
      toggles[i].dataset.pmtoggle = next;
    }
    var all = dqsa('.pmtoggleall');
    for(var i=0; i<all.length; i++) {
      all[i].dataset.pmtoggleall = next;
    }
  }
  
  function init_dropzone(){
    if(!Config.updrop) return;
    var forms = dqsa('form[action$="action=postupload"], form[action$="action=edit"]');
    for (let f of forms) {
      var div = dce('div', {className: "frame pmdropzone" });
      var label = dce('span', {textContent: Config.updrop.label+' '});
      adjab(div, label);
      adjab(f, div);
    }
    var zones = dqsa('.pmdropzone,div[data-pmdropzone]');
    for(let z of zones) mkdropzone(z);
  }
  
  function mkdropzone(Dropzone) {
    var UploadQueue = [];
    
    aE(Dropzone, 'dragenter,dragover', dragenter);
    aE(Dropzone, 'dragleave', dragleave);
    aE(Dropzone, 'drop', dragdrop);
    tap(Dropzone, clickzone);
    
    const conf = jP(Dropzone.dataset.pmdropzone, {});
    

    function findauthor(){
      if(conf.author) return conf.author;
      let authors = dqsa('input[name="author"]');
      for(let a of authors) if(a.value) return a.value;
      return '';
    }
    function dragenter(e) {
      e.preventDefault();
      Dropzone.classList.add('over');
    }
    
    function dragleave(e) {
      e.preventDefault();
      Dropzone.classList.remove('over');
    }
    
    async function dragdrop(e) {
      e.preventDefault();
      Dropzone.classList.remove('over');
      let author = findauthor();
      if(Config.updrop.areq && !author) {
        appendUpload({size: -500, name: Config.updrop.areq});
        return;
      }
      var files = event.dataTransfer.files;
      for (var i = 0; i < files.length; i++) appendUpload(files[i]);
      var pending = dqs('.pmdropzone a.uploading');
      if(!pending) await processUploads();
    }
    
    function appendUpload(file) {
      if(file.size==0) { return; }
      var sizes = Config.updrop.sizes;
      var ext = '';
      var m = file.name.match(/\.([^\.]+)$/);
      if(m) ext  = m[1].toLowerCase();
      
      var a = dce('a', {href: '#', textContent: file.name});
      
      if(file.size==-500) {
        a.className = 'error';
      }
      else if(typeof sizes[ext] == 'undefined') {
        a.className = 'error';
        a.title = Config.updrop.badtype.replace(/\#upext/, ext);
      }
      else if(file.size > sizes[ext]) {
        a.className = 'error';
        a.title = Config.updrop.toobig
        .replace(/\#upmax/, sizes[ext]).replace(/\#upext/, ext);
      }
      else {
        a.className = 'queued';
        a.pmfile = file;
      }
      Dropzone.appendChild(a);
    }
    
    async function postUpload(file) {
      var url = conf.action? conf.action : Config.updrop.action;
      var f = new FormData;
      f.append('n', conf.n? conf.n : Config.fullname);
      f.append('action', 'postupload');
      f.append('uploadfile', file);
      f.append('pmdrop', 1);
      f.append(Config.updrop.token[0], Config.updrop.token[1]);
      f.append('author', findauthor());
      try {
        var response = await fetch(url, { method: 'POST', body: f });
        if(!response.ok) {
          var msg = `HTTP error! Status: ${response.status}`;
          throw new Error(msg);
          return { error: 1, msg: msg};
        }
        var txt =  await response.text();
        var json = jP(txt, {error: 1, msg: txt });
        return json;
      }
      catch(er) {
        err(er);
        return {error: 1, msg: 'Error posting form data: ' + er.message };
      }
    }
    
    async function processUploads() {
      while(true) {
        var a = dqs('.pmdropzone a.queued');
        if(!a) return;
        a.className = 'uploading';
        var result = await postUpload(a.pmfile);
        if(result.error) {
          a.className = 'error';
          a.title = result.msg;
        }
        else {
          a.className = 'success';
          a.href = result.href;
          a.textContent = result.uprname;
          a.title = result.msg;
          delete a.pmfile;
        }
      }
    }
    
    function clickzone(e) {
      var a = e.target.closest('a');
      if(!a) return;
      e.preventDefault();
      e.stopPropagation();
      if(a.className.match(/uploading|queued|deleting/)) return;
      else if(a.className == 'success' && typeof insMarkup == 'function') {
        var fullname = conf.n? conf.n: Config.fullname;
        var pn = conf.n || e.ctrlKey? fullname + '/': '';
        var text = "Attach:"+pn+a.textContent;
        if(text.match(/\s/)) text = '[['+text+']]';
        insMarkup(function(x){
          if(x.length) return text;
          return {
            mtext: text,
            unselect: true
          };
        });
        return;
      }
      else if(a.className == 'error' || typeof insMarkup == 'undefined') {
        a.classList.add('deleting');
        setTimeout(function(){a.remove();}, 700);
        return;
      }
      // this shouldn't happen
    }
    
  }
  
  
  
  function init_draggable() {
    var els = dqsa('.draggable');
    for(var el of els) mkDrag(el);
  }
  
  function datashift() {
    var headings = dqsa('h1,h2,h3,h4,h5,h6');
    // it is common to shift next to headings, this identifies them
    for(var h of headings) h.dataset.pmcontent = h.textContent.trim();
    var data = [...dqsa('[data-shift]')];
    for(var el of data) datashiftone(el);
  }
  function datashiftone(el) {
    var x = el.dataset.shift.split(/;;+/g);
    for(var y of x) {
      var a = y.match(/^(inner:)?((?:before|after)(?:begin|end)):(.*)$/);
      if(!a) continue;
      try{
        var target = dqs(a[3]);
        if(!target) continue;
        if(el.contains(target)) {
          err("Shifted contains Target!", {el, target});
          continue;
        }
        target.insertAdjacentElement(a[2], el);
        if(a[1]) {
          var inner = [...el.childNodes];
          for(var node of inner) adjbb(el, node);
          el.remove();
          return;
        }
        el.dataset.shifted = el.dataset.shift + '';
        delete el.dataset.shift;
        delete el.dataset.unshift;
        el.style.display = el.dataset.odsp || '';
        return;
      }
      catch(e) {continue;}
    }
    
    if(el.dataset.unshift == 'hide') {
      if(!el.dataset.odsp) el.dataset.odsp = gcs(el).display;
      el.style.display = 'none';
    }
  }
  
  function wheelinputone(e) {
    const el = e.currentTarget;
    if(!(el instanceof HTMLElement)) return;
    e.preventDefault();
    var delta = Math.sign(event.deltaY);
    if(!delta) delta = -Math.sign(event.deltaX);
    if(el.nodeName === 'INPUT') {
      if(typeof el.stepUp === 'function' && typeof el.stepDown === 'function') {
        delta < 0 ? el.stepUp() : el.stepDown();
        dE(el, 'input',  { bubbles: true });
        dE(el, 'change', { bubbles: true });
      }
    } 
    else if(el.nodeName === 'SELECT') {
      const len = el.options.length;
      var n = minmax(el.selectedIndex + delta, 0, len-1);
      if(el.selectedIndex != n) {
        el.selectedIndex = n;
        dE(el, 'change', { bubbles: true });
      }
    }
  }
  function wheelinputinit(e) {
    var x = dqsa('.pmwheel,[data-pmwheel]');
    var y = [], rx = /^(number|date|month|week|range|datetime-local)$/i;
    for(var el of x) {
      if(el.nodeName === 'SELECT') y.push(el);
      else if(el.nodeName === 'INPUT' && el.type.match(rx)) y.push(el);
    }
    
    aE(y, 'wheel', wheelinputone, {passive: false});
  }
  
  function intercheckboxes(){
    let shiftKey = 0;
    let lastChecked = null;
    aE('body', 'keydown', function(e){
      if(e.key == 'Shift') shiftKey = 1;
    });
    aE('body', 'keyup', function(e){
      if(e.key == 'Shift') shiftKey = 0;
    });
    aE('body', 'change', function(e){
      const box = e.target;
      if(box.type !== 'checkbox') return;
      if(shiftKey && lastChecked) {
        const all = [...dqsa('input[type="checkbox"]')];
        const x = all.indexOf(lastChecked);
        if(x>=0) {
          const y = all.indexOf(box);
          const a = min(x, y);
          const b = max(x, y);
          const checked = box.checked;
          for (let i = a; i <= b; i++) {
            all[i].checked = checked;
          }
        }
      }
      lastChecked = box;
    });
  }
  
  function confirmUnload() {
    let forms = dqsa('form[data-confirmunload]');
    if(!forms.length) return;
    aE(forms, 'change', function(e){ this.__changed = 1; });
    aE(forms, 'submit', function(e){ this.__submitting = 1;
      setTimeout(function(){delete this.__submitting; }, 500);
    });
    
    aE(window, 'beforeunload', function(e){
      for(let f of forms) {
        let changed = (f.__changed || f.dataset.confirmunload == '2');
        if(changed && !f.__submitting) {
          e.preventDefault();
          e.returnValue = '';
          return;
        }
      }
    });
  }
  
  
  ready(function(){
    wikitext = document.getElementById('wikitext');
    var fn = [
      datashift, autotoc, inittoggle, 
      PmXMail, localTimes, 
      highlight_pre, copy_pre, 
      confirmForms, labelForms, wheelinputinit,
      init_dropzone, init_draggable,
      makesortable, intercheckboxes,
      confirmUnload
    ];
    fn.forEach(function(a){a();});
    setTimeout(datashift, 800);
  });

})(document.currentScript);

